=== modified file 'bzrlib/_dirstate_helpers_c.pyx'
--- bzrlib/_dirstate_helpers_c.pyx	2007-07-10 14:51:23 +0000
+++ bzrlib/_dirstate_helpers_c.pyx	2007-07-11 00:01:54 +0000
@@ -50,6 +50,51 @@
     char *strchr(char *s1, char c)
     int strncmp(char *s1, char *s2, int len)
     int strcmp(char *s1, char *s2)
+    void *memchr(void *s, int c, size_t len)
+    int memcmp(void *b1, void *b2, size_t len)
+    # ??? memrchr is a GNU extension :(
+    # void *memrchr(void *s, int c, size_t len)
+
+
+cdef void* _my_memrchr(void *s, int c, size_t n):
+    # memrchr seems to be a GNU extension, so we have to implement it ourselves
+    # We semi-cheat, and just call memchr repeatedly until we run out of path.
+    # And then return the last match.
+    cdef void *pos
+    cdef void *end
+    cdef void *last_pos
+
+    last_pos = NULL
+    end = <void*>((<char*>s) + n)
+    pos = memchr(s, c, n)
+    while pos != NULL and pos < end:
+        last_pos = pos
+        pos = memchr(pos+1, c, end-pos)
+    return last_pos
+
+
+def _py_memrchr(s, c):
+    """Just to expose _my_memrchr for testing.
+
+    :param s: The Python string to search
+    :param c: The character to search for
+    :return: The offset to the last instance of 'c' in s
+    """
+    cdef void *_s
+    cdef void *found
+    cdef int length
+    cdef char *_c
+
+    _s = PyString_AsString(s)
+    length = PyString_Size(s)
+
+    _c = PyString_AsString(c)
+    assert PyString_Size(c) == 1,\
+        'Must be a single character string, not %s' % (c,)
+    found = _my_memrchr(_s, _c[0], length)
+    if found == NULL:
+        return None
+    return found - _s
 
 
 cdef int _cmp_by_dirs(char *path1, int size1, char *path2, int size2):
@@ -135,6 +180,202 @@
                         PyString_Size(path2))
 
 
+def cmp_path_by_dirblock_c(path1, path2):
+    """Compare two paths based on what directory they are in.
+
+    This generates a sort order, such that all children of a directory are
+    sorted together, and grandchildren are in the same order as the
+    children appear. But all grandchildren come after all children.
+
+    :param path1: first path
+    :param path2: the second path
+    :return: positive number if ``path1`` comes first,
+        0 if paths are equal
+        and a negative number if ``path2`` sorts first
+    """
+    return _cmp_path_by_dirblock(PyString_AsString(path1),
+                                 PyString_Size(path1),
+                                 PyString_AsString(path2),
+                                 PyString_Size(path2))
+
+
+cdef _cmp_path_by_dirblock(char *path1, int path1_len,
+                           char *path2, int path2_len):
+    cdef char *dirname1
+    cdef int dirname1_len
+    cdef char *dirname2
+    cdef int dirname2_len
+    cdef char *basename1
+    cdef int basename1_len
+    cdef char *basename2
+    cdef int basename2_len
+    cdef int cur_len
+    cdef int cmp_val
+
+    dirname1 = path1
+    dirname1_len = path1_len
+
+    dirname2 = path2
+    dirname2_len = path2_len
+
+    if path1_len == 0 and path2_len == 0:
+        return 0
+
+    if path1_len == 0:
+        return -1
+
+    if path2_len == 0:
+        return 1
+
+    basename1 = <char*>_my_memrchr(dirname1, c'/', dirname1_len)
+
+    if basename1 == NULL:
+        basename1 = dirname1
+        dirname1 = ''
+        dirname1_len = 0
+    else:
+        cur_len = basename1 - dirname1
+        basename1 = basename1 + 1
+        basename1_len = dirname1_len - cur_len - 1
+        dirname1_len = cur_len
+
+    basename2 = <char*>_my_memrchr(dirname2, c'/', dirname2_len)
+
+    if basename2 == NULL:
+        basename2 = dirname2
+        dirname2 = ''
+        dirname2_len = 0
+    else:
+        cur_len = basename2 - dirname2
+        basename2 = basename2 + 1
+        basename2_len = dirname2_len - cur_len - 1
+        dirname2_len = cur_len
+
+    cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
+                           dirname2, dirname2_len)
+    if cmp_val != 0:
+        return cmp_val
+
+    cur_len = basename1_len
+    if basename2_len < basename1_len:
+        cur_len = basename2_len
+
+    cmp_val = memcmp(basename1, basename2, cur_len)
+    if cmp_val != 0:
+        return cmp_val
+    if basename1_len == basename2_len:
+        return 0
+    if basename1_len < basename2_len:
+        return -1
+    return 1
+
+
+def bisect_path_left_c(paths, path):
+    """Return the index where to insert path into paths.
+
+    This uses a path-wise comparison so we get::
+        a
+        a-b
+        a=b
+        a/b
+    Rather than::
+        a
+        a-b
+        a/b
+        a=b
+    :param paths: A list of paths to search through
+    :param path: A single path to insert
+    :return: An offset where 'path' can be inserted.
+    :seealso: bisect.bisect_left
+    """
+    cdef int _lo
+    cdef int _hi
+    cdef int _mid
+    cdef char *path_str
+    cdef int path_size
+    cdef char *cur_str
+    cdef int cur_size
+    cdef void *cur
+
+    if not PyList_CheckExact(paths):
+        raise TypeError('you must pass a python list for paths')
+    if not PyString_CheckExact(path):
+        raise TypeError('you must pass a string for path')
+
+    _hi = len(paths)
+    _lo = 0
+
+    path_str = PyString_AsString(path)
+    path_size = PyString_Size(path)
+
+    while _lo < _hi:
+        _mid = (_lo + _hi) / 2
+        cur = PyList_GetItem_object_void(paths, _mid)
+        cur_str = PyString_AS_STRING_void(cur)
+        cur_size = PyString_GET_SIZE_void(cur)
+        if _cmp_path_by_dirblock(cur_str, cur_size, path_str, path_size) < 0:
+            _lo = _mid + 1
+        else:
+            _hi = _mid
+    return _lo
+
+
+def bisect_path_right_c(paths, path):
+    """Return the index where to insert path into paths.
+
+    This uses a path-wise comparison so we get::
+        a
+        a-b
+        a=b
+        a/b
+    Rather than::
+        a
+        a-b
+        a/b
+        a=b
+    :param paths: A list of paths to search through
+    :param path: A single path to insert
+    :return: An offset where 'path' can be inserted.
+    :seealso: bisect.bisect_right
+    """
+    cdef int _lo
+    cdef int _hi
+    cdef int _mid
+    cdef char *path_str
+    cdef int path_size
+    cdef char *cur_str
+    cdef int cur_size
+    cdef void *cur
+    cdef int cmp_val
+
+    if not PyList_CheckExact(paths):
+        raise TypeError('you must pass a python list for paths')
+    if not PyString_CheckExact(path):
+        raise TypeError('you must pass a string for path')
+
+    _hi = len(paths)
+    _lo = 0
+
+    path_str = PyString_AsString(path)
+    path_size = PyString_Size(path)
+
+    while _lo < _hi:
+        _mid = (_lo + _hi) / 2
+        cur = PyList_GetItem_object_void(paths, _mid)
+        cur_str = PyString_AS_STRING_void(cur)
+        cur_size = PyString_GET_SIZE_void(cur)
+        cmp_val = _cmp_path_by_dirblock(path_str, path_size, cur_str, cur_size)
+        print 'c_left mid: %d, cmp_val %d, cur_str %r, path_str %r' % (
+            _mid, cmp_val,
+            PyString_FromStringAndSize(cur_str, cur_size),
+            PyString_FromStringAndSize(path_str, path_size))
+        if cmp_val < 0:
+            _hi = _mid
+        else:
+            _lo = _mid + 1
+    return _lo
+
+
 def bisect_dirblock_c(dirblocks, dirname, lo=0, hi=None, cache=None):
     """Return the index where to insert dirname into the dirblocks.
 
@@ -168,7 +409,7 @@
     dirname_size = PyString_Size(dirname)
 
     while _lo < _hi:
-        _mid = (_lo+_hi)/2
+        _mid = (_lo + _hi) / 2
         # Grab the dirname for the current dirblock
         # cur = dirblocks[_mid][0]
         cur = PyTuple_GetItem_void_void(
@@ -176,7 +417,7 @@
         cur_str = PyString_AS_STRING_void(cur)
         cur_size = PyString_GET_SIZE_void(cur)
         if _cmp_by_dirs(cur_str, cur_size, dirname_str, dirname_size) < 0:
-            _lo = _mid+1
+            _lo = _mid + 1
         else:
             _hi = _mid
     return _lo
@@ -204,6 +445,7 @@
         """Return a pointer to the start of the next field."""
         cdef char *next
         next = self.cur
+        # XXX: Change this to not use 'memchr' instead of 'strchr'
         self.cur = strchr(next, c'\0')
         size[0] = self.cur - next
         self.cur = self.cur + 1

=== modified file 'bzrlib/_dirstate_helpers_py.py'
--- bzrlib/_dirstate_helpers_py.py	2007-07-10 14:51:23 +0000
+++ bzrlib/_dirstate_helpers_py.py	2007-07-11 00:01:54 +0000
@@ -16,6 +16,88 @@
 
 """Python implementations of Dirstate Helper functions."""
 
+import os
+
+
+def bisect_path_left_py(paths, path):
+    """Return the index where to insert path into paths.
+
+    This uses the dirblock sorting. So all children in a directory come before
+    the children of children. For example::
+
+        a/
+          b/
+            c
+          d/
+            e
+          b-c
+          d-e
+        a-a
+        a=c
+
+    Will be sorted as::
+
+        a
+        a-a
+        a=c
+        a/b
+        a/b-c
+        a/d
+        a/d-e
+        a/b/c
+        a/d/e
+
+    :param paths: A list of paths to search through
+    :param path: A single path to insert
+    :return: An offset where 'path' can be inserted.
+    :seealso: bisect.bisect_left
+    """
+    hi = len(paths)
+    lo = 0
+    while lo < hi:
+        mid = (lo + hi) // 2
+        # Grab the dirname for the current dirblock
+        cur = paths[mid]
+        cmp_val = cmp_path_by_dirblock_py(cur, path)
+        print 'py_left mid: %d, cmp_val %d, cur_str %r, path_str %r' % (
+            mid, cmp_val, cur, path)
+        if cmp_val < 0:
+            lo = mid + 1
+        else:
+            hi = mid
+    return lo
+
+
+def bisect_path_right_py(paths, path):
+    """Return the index where to insert path into paths.
+
+    This uses a path-wise comparison so we get::
+        a
+        a-b
+        a=b
+        a/b
+    Rather than::
+        a
+        a-b
+        a/b
+        a=b
+    :param paths: A list of paths to search through
+    :param path: A single path to insert
+    :return: An offset where 'path' can be inserted.
+    :seealso: bisect.bisect_right
+    """
+    hi = len(paths)
+    lo = 0
+    while lo < hi:
+        mid = (lo+hi)//2
+        # Grab the dirname for the current dirblock
+        cur = paths[mid]
+        if cmp_path_by_dirblock_py(path, cur) < 0:
+            hi = mid
+        else:
+            lo = mid + 1
+    return lo
+
 
 def bisect_dirblock_py(dirblocks, dirname, lo=0, hi=None, cache={}):
     """Return the index where to insert dirname into the dirblocks.
@@ -35,7 +117,7 @@
         dirname_split = dirname.split('/')
         cache[dirname] = dirname_split
     while lo < hi:
-        mid = (lo+hi)//2
+        mid = (lo + hi) // 2
         # Grab the dirname for the current dirblock
         cur = dirblocks[mid][0]
         try:
@@ -43,7 +125,7 @@
         except KeyError:
             cur_split = cur.split('/')
             cache[cur] = cur_split
-        if cur_split < dirname_split: lo = mid+1
+        if cur_split < dirname_split: lo = mid + 1
         else: hi = mid
     return lo
 
@@ -169,3 +251,23 @@
         and negative number if ``path2`` sorts first
     """
     return cmp(path1.split('/'), path2.split('/'))
+
+
+def cmp_path_by_dirblock_py(path1, path2):
+    """Compare two paths based on what directory they are in.
+
+    This generates a sort order, such that all children of a directory are
+    sorted together, and grandchildren are in the same order as the
+    children appear. But all grandchildren come after all children.
+
+    :param path1: first path
+    :param path2: the second path
+    :return: positive number if ``path1`` comes first,
+        0 if paths are equal
+        and a negative number if ``path2`` sorts first
+    """
+    dirname1, basename1 = os.path.split(path1)
+    key1 = (dirname1.split('/'), basename1)
+    dirname2, basename2 = os.path.split(path2)
+    key2 = (dirname2.split('/'), basename2)
+    return cmp(key1, key2)

=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2007-07-10 14:51:23 +0000
+++ bzrlib/dirstate.py	2007-07-11 00:01:54 +0000
@@ -199,7 +199,6 @@
 
 """
 
-
 import bisect
 import binascii
 import errno
@@ -423,12 +422,14 @@
         if self._id_index:
             self._id_index.setdefault(entry_key[2], set()).add(entry_key)
 
-    def _bisect(self, dir_name_list):
+    def _bisect(self, paths):
         """Bisect through the disk structure for specific rows.
 
-        :param dir_name_list: A list of (dir, name) pairs.
-        :return: A dict mapping (dir, name) => entry for found entries. Missing
+        :param paths: A list of paths to find
+        :return: A dict mapping path => entries for found entries. Missing
                  entries will not be in the map.
+                 The list is not sorted, and entries will be populated
+                 based on when they were read.
         """
         self._requires_lock()
         # We need the file pointer to be right after the initial header block
@@ -454,7 +455,7 @@
         found = {}
 
         # Avoid infinite seeking
-        max_count = 30*len(dir_name_list)
+        max_count = 30*len(paths)
         count = 0
         # pending is a list of places to look.
         # each entry is a tuple of low, high, dir_names
@@ -462,7 +463,7 @@
         #   high -> the last byte offset (inclusive)
         #   dir_names -> The list of (dir, name) pairs that should be found in
         #                the [low, high] range
-        pending = [(low, high, dir_name_list)]
+        pending = [(low, high, paths)]
 
         page_size = self._bisect_page_size
 
@@ -521,8 +522,11 @@
                 # Find what entries we are looking for, which occur before and
                 # after this first record.
                 after = start
-                first_dir_name = (first_fields[1], first_fields[2])
-                first_loc = bisect.bisect_left(cur_files, first_dir_name)
+                if first_fields[1]:
+                    first_path = first_fields[1] + '/' + first_fields[2]
+                else:
+                    first_path = first_fields[2]
+                first_loc = bisect_path_left(cur_files, first_path)
 
                 # These exist before the current location
                 pre = cur_files[:first_loc]
@@ -545,8 +549,11 @@
                 else:
                     after = mid + len(block)
 
-                last_dir_name = (last_fields[1], last_fields[2])
-                last_loc = bisect.bisect_right(post, last_dir_name)
+                if last_fields[1]:
+                    last_path = last_fields[1] + '/' + last_fields[2]
+                else:
+                    last_path = last_fields[2]
+                last_loc = bisect_path_right(post, last_path)
 
                 middle_files = post[:last_loc]
                 post = post[last_loc:]
@@ -557,33 +564,36 @@
                     # Either we will find them here, or we can mark them as
                     # missing.
 
-                    if middle_files[0] == first_dir_name:
+                    if middle_files[0] == first_path:
                         # We might need to go before this location
-                        pre.append(first_dir_name)
-                    if middle_files[-1] == last_dir_name:
-                        post.insert(0, last_dir_name)
+                        pre.append(first_path)
+                    if middle_files[-1] == last_path:
+                        post.insert(0, last_path)
 
                     # Find out what paths we have
-                    paths = {first_dir_name:[first_fields]}
-                    # last_dir_name might == first_dir_name so we need to be
+                    paths = {first_path:[first_fields]}
+                    # last_path might == first_path so we need to be
                     # careful if we should append rather than overwrite
                     if last_entry_num != first_entry_num:
-                        paths.setdefault(last_dir_name, []).append(last_fields)
+                        paths.setdefault(last_path, []).append(last_fields)
                     for num in xrange(first_entry_num+1, last_entry_num):
                         # TODO: jam 20070223 We are already splitting here, so
                         #       shouldn't we just split the whole thing rather
                         #       than doing the split again in add_one_record?
                         fields = entries[num].split('\0')
-                        dir_name = (fields[1], fields[2])
-                        paths.setdefault(dir_name, []).append(fields)
+                        if fields[1]:
+                            path = fields[1] + '/' + fields[2]
+                        else:
+                            path = fields[2]
+                        paths.setdefault(path, []).append(fields)
 
-                    for dir_name in middle_files:
-                        for fields in paths.get(dir_name, []):
+                    for path in middle_files:
+                        for fields in paths.get(path, []):
                             # offset by 1 because of the opening '\0'
                             # consider changing fields_to_entry to avoid the
                             # extra list slice
                             entry = fields_to_entry(fields[1:])
-                            found.setdefault(dir_name, []).append(entry)
+                            found.setdefault(path, []).append(entry)
 
             # Now we have split up everything into pre, middle, and post, and
             # we have handled everything that fell in 'middle'.
@@ -705,7 +715,7 @@
                 # after this first record.
                 after = start
                 first_dir = first_fields[1]
-                first_loc = bisect.bisect_left(cur_dirs, first_dir)
+                first_loc = bisect_path_left(cur_dirs, first_dir)
 
                 # These exist before the current location
                 pre = cur_dirs[:first_loc]
@@ -729,7 +739,7 @@
                     after = mid + len(block)
 
                 last_dir = last_fields[1]
-                last_loc = bisect.bisect_right(post, last_dir)
+                last_loc = bisect_path_right(post, last_dir)
 
                 middle_files = post[:last_loc]
                 post = post[last_loc:]
@@ -779,7 +789,7 @@
 
         return found
 
-    def _bisect_recursive(self, dir_name_list):
+    def _bisect_recursive(self, paths):
         """Bisect for entries for all paths and their children.
 
         This will use bisect to find all records for the supplied paths. It
@@ -798,7 +808,7 @@
         # Directories that have been read
         processed_dirs = set()
         # Get the ball rolling with the first bisect for all entries.
-        newly_found = self._bisect(dir_name_list)
+        newly_found = self._bisect(paths)
 
         while newly_found:
             # Directories that need to be read
@@ -2267,11 +2277,15 @@
     from bzrlib._dirstate_helpers_c import (
         _read_dirblocks_c as _read_dirblocks,
         bisect_dirblock_c as bisect_dirblock,
+        bisect_path_left_c as bisect_path_left,
+        bisect_path_right_c as bisect_path_right,
         cmp_by_dirs_c as cmp_by_dirs,
         )
 except ImportError:
     from bzrlib._dirstate_helpers_py import (
         _read_dirblocks_py as _read_dirblocks,
         bisect_dirblock_py as bisect_dirblock,
+        bisect_path_left_py as bisect_path_left,
+        bisect_path_right_py as bisect_path_right,
         cmp_by_dirs_py as cmp_by_dirs,
         )

=== modified file 'bzrlib/tests/test__dirstate_helpers.py'
--- bzrlib/tests/test__dirstate_helpers.py	2007-07-10 14:51:23 +0000
+++ bzrlib/tests/test__dirstate_helpers.py	2007-07-11 00:01:54 +0000
@@ -17,6 +17,7 @@
 """Tests for the compiled dirstate helpers."""
 
 import bisect
+import os
 
 from bzrlib import (
     tests,
@@ -37,6 +38,313 @@
 CompiledDirstateHelpersFeature = _CompiledDirstateHelpersFeature()
 
 
+class TestBisectPathMixin(object):
+    """Test that bisect_path_*() returns the expected values.
+
+    bisect_path_* is intended to work like bisect.bisect_*() except it
+    knows it is working on paths that are sorted by ('path', 'to', 'foo')
+    chunks rather than by raw 'path/to/foo'.
+    """
+
+    def get_bisect_path(self):
+        """Return an implementation of bisect_path_*"""
+        raise NotImplementedError
+
+    def get_bisect(self):
+        """Return a version of bisect.bisect_*.
+
+        Also, for the 'exists' check, return the offset to the real values.
+        For example bisect_left returns the index of an entry, while
+        bisect_right returns the index *after* an entry
+
+        :return: (bisect_func, offset)
+        """
+        raise NotImplementedError
+
+    def assertBisect(self, paths, split_paths, path, exists=True):
+        """Assert that bisect_split works like bisect_left on the split paths.
+
+        :param paths: A list of path names
+        :param split_paths: A list of path names that are already split up by directory
+            ('path/to/foo' => ('path', 'to', 'foo'))
+        :param path: The path we are indexing.
+        :param exists: The path should be present, so make sure the
+            final location actually points to the right value.
+
+        All other arguments will be passed along.
+        """
+        bisect_path = self.get_bisect_path()
+        self.assertIsInstance(paths, list)
+        bisect_path_idx = bisect_path(paths, path)
+        split_path = self.split_for_dirblocks([path])[0]
+        bisect_func, offset = self.get_bisect()
+        bisect_split_idx = bisect_func(split_paths, split_path)
+        self.assertEqual(bisect_split_idx, bisect_path_idx,
+                         '%s disagreed. %s != %s'
+                         ' for key %r'
+                         % (bisect_path.__name__,
+                            bisect_split_idx, bisect_path_idx, path)
+                         )
+        if exists:
+            self.assertEqual(path, paths[bisect_path_idx+offset])
+
+    def split_for_dirblocks(self, paths):
+        dir_split_paths = []
+        for path in paths:
+            dirname, basename = os.path.split(path)
+            dir_split_paths.append((dirname.split('/'), basename))
+        dir_split_paths.sort()
+        return dir_split_paths
+
+    def test_simple(self):
+        """In the simple case it works just like bisect_left"""
+        paths = ['', 'a', 'b', 'c', 'd']
+        split_paths = self.split_for_dirblocks(paths)
+        for path in paths:
+            self.assertBisect(paths, split_paths, path, exists=True)
+        self.assertBisect(paths, split_paths, '_', exists=False)
+        self.assertBisect(paths, split_paths, 'aa', exists=False)
+        self.assertBisect(paths, split_paths, 'bb', exists=False)
+        self.assertBisect(paths, split_paths, 'cc', exists=False)
+        self.assertBisect(paths, split_paths, 'dd', exists=False)
+        self.assertBisect(paths, split_paths, 'a/a', exists=False)
+        self.assertBisect(paths, split_paths, 'b/b', exists=False)
+        self.assertBisect(paths, split_paths, 'c/c', exists=False)
+        self.assertBisect(paths, split_paths, 'd/d', exists=False)
+
+    def test_involved(self):
+        """This is where bisect_path_* diverges slightly."""
+        # This is the list of paths and their contents
+        # a/
+        #   a/
+        #     a
+        #     z
+        #   a-a/
+        #     a
+        #   a-z/
+        #     z
+        #   a=a/
+        #     a
+        #   a=z/
+        #     z
+        #   z/
+        #     a
+        #     z
+        #   z-a
+        #   z-z
+        #   z=a
+        #   z=z
+        # a-a/
+        #   a
+        # a-z/
+        #   z
+        # a=a/
+        #   a
+        # a=z/
+        #   z
+        # This is the exact order that is stored by dirstate
+        # All children in a directory are mentioned before an children of
+        # children are mentioned.
+        # So all the root-directory paths, then all the
+        # first sub directory, etc.
+        paths = [# content of '/'
+                 '', 'a', 'a-a', 'a-z', 'a=a', 'a=z',
+                 # content of 'a/'
+                 'a/a', 'a/a-a', 'a/a-z',
+                 'a/a=a', 'a/a=z',
+                 'a/z', 'a/z-a', 'a/z-z',
+                 'a/z=a', 'a/z=z',
+                 # content of 'a/a/'
+                 'a/a/a', 'a/a/z',
+                 # content of 'a/a-a'
+                 'a/a-a/a',
+                 # content of 'a/a-z'
+                 'a/a-z/z',
+                 # content of 'a/a=a'
+                 'a/a=a/a',
+                 # content of 'a/a=z'
+                 'a/a=z/z',
+                 # content of 'a/z/'
+                 'a/z/a', 'a/z/z',
+                 # content of 'a-a'
+                 'a-a/a',
+                 # content of 'a-z'
+                 'a-z/z',
+                 # content of 'a=a'
+                 'a=a/a',
+                 # content of 'a=z'
+                 'a=z/z',
+                ]
+        split_paths = self.split_for_dirblocks(paths)
+        sorted_paths = []
+        for dir_parts, basename in split_paths:
+            if dir_parts == ['']:
+                sorted_paths.append(basename)
+            else:
+                sorted_paths.append('/'.join(dir_parts + [basename]))
+
+        self.assertEqual(sorted_paths, paths)
+
+        for path in paths:
+            self.assertBisect(paths, split_paths, path, exists=True)
+
+
+class TestBisectPathLeft(tests.TestCase, TestBisectPathMixin):
+
+    def get_bisect_path(self):
+        from bzrlib._dirstate_helpers_py import bisect_path_left_py
+        return bisect_path_left_py
+
+    def get_bisect(self):
+        return bisect.bisect_left, 0
+
+
+class TestCompiledBisectPathLeft(TestBisectPathLeft):
+    """Test that bisect_path_left() returns the expected values.
+
+    This runs all the normal tests that TestBisectDirblock did, but uses the
+    compiled version.
+    """
+
+    _test_needs_features = [CompiledDirstateHelpersFeature]
+
+    def get_bisect_path(self):
+        from bzrlib._dirstate_helpers_c import bisect_path_left_c
+        return bisect_path_left_c
+
+
+class TestBisectPathRight(tests.TestCase, TestBisectPathMixin):
+
+    def get_bisect_path(self):
+        from bzrlib._dirstate_helpers_py import bisect_path_right_py
+        return bisect_path_right_py
+
+    def get_bisect(self):
+        return bisect.bisect_right, -1
+
+
+class TestCompiledBisectPathRight(TestBisectPathRight):
+    """Test that bisect_path_right() returns the expected values.
+
+    This runs all the normal tests that TestBisectDirblock did, but uses the
+    compiled version.
+    """
+
+    _test_needs_features = [CompiledDirstateHelpersFeature]
+
+    def get_bisect_path(self):
+        from bzrlib._dirstate_helpers_c import bisect_path_right_c
+        return bisect_path_right_c
+
+
+class TestBisectDirblock(tests.TestCase):
+    """Test that bisect_dirblock() returns the expected values.
+
+    bisect_dirblock is intended to work like bisect.bisect_left() except it
+    knows it is working on dirblocks and that dirblocks are sorted by ('path',
+    'to', 'foo') chunks rather than by raw 'path/to/foo'.
+    """
+
+    def get_bisect_dirblock(self):
+        """Return an implementation of bisect_dirblock"""
+        from bzrlib._dirstate_helpers_py import bisect_dirblock_py
+        return bisect_dirblock_py
+
+    def assertBisect(self, dirblocks, split_dirblocks, path, *args, **kwargs):
+        """Assert that bisect_split works like bisect_left on the split paths.
+
+        :param dirblocks: A list of (path, [info]) pairs.
+        :param split_dirblocks: A list of ((split, path), [info]) pairs.
+        :param path: The path we are indexing.
+
+        All other arguments will be passed along.
+        """
+        bisect_dirblock = self.get_bisect_dirblock()
+        self.assertIsInstance(dirblocks, list)
+        bisect_split_idx = bisect_dirblock(dirblocks, path, *args, **kwargs)
+        split_dirblock = (path.split('/'), [])
+        bisect_left_idx = bisect.bisect_left(split_dirblocks, split_dirblock,
+                                             *args)
+        self.assertEqual(bisect_left_idx, bisect_split_idx,
+                         'bisect_split disagreed. %s != %s'
+                         ' for key %r'
+                         % (bisect_left_idx, bisect_split_idx, path)
+                         )
+
+    def paths_to_dirblocks(self, paths):
+        """Convert a list of paths into dirblock form.
+
+        Also, ensure that the paths are in proper sorted order.
+        """
+        dirblocks = [(path, []) for path in paths]
+        split_dirblocks = [(path.split('/'), []) for path in paths]
+        self.assertEqual(sorted(split_dirblocks), split_dirblocks)
+        return dirblocks, split_dirblocks
+
+    def test_simple(self):
+        """In the simple case it works just like bisect_left"""
+        paths = ['', 'a', 'b', 'c', 'd']
+        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
+        for path in paths:
+            self.assertBisect(dirblocks, split_dirblocks, path)
+        self.assertBisect(dirblocks, split_dirblocks, '_')
+        self.assertBisect(dirblocks, split_dirblocks, 'aa')
+        self.assertBisect(dirblocks, split_dirblocks, 'bb')
+        self.assertBisect(dirblocks, split_dirblocks, 'cc')
+        self.assertBisect(dirblocks, split_dirblocks, 'dd')
+        self.assertBisect(dirblocks, split_dirblocks, 'a/a')
+        self.assertBisect(dirblocks, split_dirblocks, 'b/b')
+        self.assertBisect(dirblocks, split_dirblocks, 'c/c')
+        self.assertBisect(dirblocks, split_dirblocks, 'd/d')
+
+    def test_involved(self):
+        """This is where bisect_left diverges slightly."""
+        paths = ['', 'a',
+                 'a/a', 'a/a/a', 'a/a/z', 'a/a-a', 'a/a-z',
+                 'a/z', 'a/z/a', 'a/z/z', 'a/z-a', 'a/z-z',
+                 'a-a', 'a-z',
+                 'z', 'z/a/a', 'z/a/z', 'z/a-a', 'z/a-z',
+                 'z/z', 'z/z/a', 'z/z/z', 'z/z-a', 'z/z-z',
+                 'z-a', 'z-z',
+                ]
+        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
+        for path in paths:
+            self.assertBisect(dirblocks, split_dirblocks, path)
+
+    def test_involved_cached(self):
+        """This is where bisect_left diverges slightly."""
+        paths = ['', 'a',
+                 'a/a', 'a/a/a', 'a/a/z', 'a/a-a', 'a/a-z',
+                 'a/z', 'a/z/a', 'a/z/z', 'a/z-a', 'a/z-z',
+                 'a-a', 'a-z',
+                 'z', 'z/a/a', 'z/a/z', 'z/a-a', 'z/a-z',
+                 'z/z', 'z/z/a', 'z/z/z', 'z/z-a', 'z/z-z',
+                 'z-a', 'z-z',
+                ]
+        cache = {}
+        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
+        for path in paths:
+            self.assertBisect(dirblocks, split_dirblocks, path, cache=cache)
+
+
+class TestCompiledBisectDirblock(TestBisectDirblock):
+    """Test that bisect_dirblock() returns the expected values.
+
+    bisect_dirblock is intended to work like bisect.bisect_left() except it
+    knows it is working on dirblocks and that dirblocks are sorted by ('path',
+    'to', 'foo') chunks rather than by raw 'path/to/foo'.
+
+    This runs all the normal tests that TestBisectDirblock did, but uses the
+    compiled version.
+    """
+
+    _test_needs_features = [CompiledDirstateHelpersFeature]
+
+    def get_bisect_dirblock(self):
+        from bzrlib._dirstate_helpers_c import bisect_dirblock_c
+        return bisect_dirblock_c
+
+
 class TestCmpByDirs(tests.TestCase):
 
     def get_cmp_by_dirs(self):
@@ -136,8 +444,8 @@
         self.assertCmpByDirs(-1, 'ab/cd', 'ab-cd')
 
 
-class TestCCmpByDirs(TestCmpByDirs):
-    """Test the C implementation of cmp_by_dirs"""
+class TestCompiledCmpByDirs(TestCmpByDirs):
+    """Test the pyrex implementation of cmp_by_dirs"""
 
     _test_needs_features = [CompiledDirstateHelpersFeature]
 
@@ -146,109 +454,160 @@
         return cmp_by_dirs_c
 
 
-class TestBisectDirblock(tests.TestCase):
-    """Test that bisect_dirblock() returns the expected values.
-
-    bisect_dirblock is intended to work like bisect.bisect_left() except it
-    knows it is working on dirblocks and that dirblocks are sorted by ('path',
-    'to', 'foo') chunks rather than by raw 'path/to/foo'.
-    """
-
-    def get_bisect_dirblock(self):
-        """Return an implementation of bisect_dirblock"""
-        from bzrlib._dirstate_helpers_py import bisect_dirblock_py
-        return bisect_dirblock_py
-
-    def assertBisect(self, dirblocks, split_dirblocks, path, *args, **kwargs):
-        """Assert that bisect_split works like bisect_left on the split paths.
-
-        :param dirblocks: A list of (path, [info]) pairs.
-        :param split_dirblocks: A list of ((split, path), [info]) pairs.
-        :param path: The path we are indexing.
-
-        All other arguments will be passed along.
-        """
-        bisect_dirblock = self.get_bisect_dirblock()
-        self.assertIsInstance(dirblocks, list)
-        bisect_split_idx = bisect_dirblock(dirblocks, path, *args, **kwargs)
-        split_dirblock = (path.split('/'), [])
-        bisect_left_idx = bisect.bisect_left(split_dirblocks, split_dirblock,
-                                             *args)
-        self.assertEqual(bisect_left_idx, bisect_split_idx,
-                         'bisect_split disagreed. %s != %s'
-                         ' for key %s'
-                         % (bisect_left_idx, bisect_split_idx, path)
-                         )
-
-    def paths_to_dirblocks(self, paths):
-        """Convert a list of paths into dirblock form.
-
-        Also, ensure that the paths are in proper sorted order.
-        """
-        dirblocks = [(path, []) for path in paths]
-        split_dirblocks = [(path.split('/'), []) for path in paths]
-        self.assertEqual(sorted(split_dirblocks), split_dirblocks)
-        return dirblocks, split_dirblocks
-
-    def test_simple(self):
-        """In the simple case it works just like bisect_left"""
-        paths = ['', 'a', 'b', 'c', 'd']
-        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
-        for path in paths:
-            self.assertBisect(dirblocks, split_dirblocks, path)
-        self.assertBisect(dirblocks, split_dirblocks, '_')
-        self.assertBisect(dirblocks, split_dirblocks, 'aa')
-        self.assertBisect(dirblocks, split_dirblocks, 'bb')
-        self.assertBisect(dirblocks, split_dirblocks, 'cc')
-        self.assertBisect(dirblocks, split_dirblocks, 'dd')
-        self.assertBisect(dirblocks, split_dirblocks, 'a/a')
-        self.assertBisect(dirblocks, split_dirblocks, 'b/b')
-        self.assertBisect(dirblocks, split_dirblocks, 'c/c')
-        self.assertBisect(dirblocks, split_dirblocks, 'd/d')
-
-    def test_involved(self):
-        """This is where bisect_left diverges slightly."""
-        paths = ['', 'a',
-                 'a/a', 'a/a/a', 'a/a/z', 'a/a-a', 'a/a-z',
-                 'a/z', 'a/z/a', 'a/z/z', 'a/z-a', 'a/z-z',
-                 'a-a', 'a-z',
-                 'z', 'z/a/a', 'z/a/z', 'z/a-a', 'z/a-z',
-                 'z/z', 'z/z/a', 'z/z/z', 'z/z-a', 'z/z-z',
-                 'z-a', 'z-z',
-                ]
-        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
-        for path in paths:
-            self.assertBisect(dirblocks, split_dirblocks, path)
-
-    def test_involved_cached(self):
-        """This is where bisect_left diverges slightly."""
-        paths = ['', 'a',
-                 'a/a', 'a/a/a', 'a/a/z', 'a/a-a', 'a/a-z',
-                 'a/z', 'a/z/a', 'a/z/z', 'a/z-a', 'a/z-z',
-                 'a-a', 'a-z',
-                 'z', 'z/a/a', 'z/a/z', 'z/a-a', 'z/a-z',
-                 'z/z', 'z/z/a', 'z/z/z', 'z/z-a', 'z/z-z',
-                 'z-a', 'z-z',
-                ]
-        cache = {}
-        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
-        for path in paths:
-            self.assertBisect(dirblocks, split_dirblocks, path, cache=cache)
-
-
-class TestCompiledBisectDirblock(TestBisectDirblock):
-    """Test that bisect_dirblock() returns the expected values.
-
-    bisect_dirblock is intended to work like bisect.bisect_left() except it
-    knows it is working on dirblocks and that dirblocks are sorted by ('path',
-    'to', 'foo') chunks rather than by raw 'path/to/foo'.
-
-    This runs all the normal tests that TestBisectDirblock did, but uses the
-    compiled version.
-    """
-
-    _test_needs_features = [CompiledDirstateHelpersFeature]
-
-    def get_bisect_dirblock(self):
-        from bzrlib._dirstate_helpers_c import bisect_dirblock_c
-        return bisect_dirblock_c
+class TestCmpPathByDirblock(tests.TestCase):
+
+    def get_cmp_path_by_dirblock(self):
+        """Get a specific implementation of cmp_path_by_dirblock."""
+        from bzrlib._dirstate_helpers_py import cmp_path_by_dirblock_py
+        return cmp_path_by_dirblock_py
+
+    def assertCmpPathByDirblock(self, paths):
+        """Compare all paths and make sure they evaluate to the correct order.
+
+        This does N^2 comparisons. It is assumed that ``paths`` is properly
+        sorted list.
+
+        :param paths: a sorted list of paths to compare
+        """
+        # First, make sure the paths being passed in are correct
+        def _key(p):
+            dirname, basename = os.path.split(p)
+            return dirname.split('/'), basename
+        self.assertEqual(sorted(paths, key=_key), paths)
+
+        cmp_path_by_dirblock = self.get_cmp_path_by_dirblock()
+        for idx1, path1 in enumerate(paths):
+            for idx2, path2 in enumerate(paths):
+                cmp_val = cmp_path_by_dirblock(path1, path2)
+                if idx1 < idx2:
+                    self.assertTrue(cmp_val < 0,
+                        '%s did not state that %r came before %r, cmp=%s'
+                        % (cmp_path_by_dirblock.__name__,
+                           path1, path2, cmp_val))
+                elif idx1 > idx2:
+                    self.assertTrue(cmp_val > 0,
+                        '%s did not state that %r came after %r, cmp=%s'
+                        % (cmp_path_by_dirblock.__name__,
+                           path1, path2, cmp_val))
+                else: # idx1 == idx2
+                    self.assertTrue(cmp_val == 0,
+                        '%s did not state that %r == %r, cmp=%s'
+                        % (cmp_path_by_dirblock.__name__,
+                           path1, path2, cmp_val))
+
+    def test_cmp_simple_paths(self):
+        """Compare against the empty string."""
+        self.assertCmpPathByDirblock(['', 'a', 'ab', 'abc', 'a/b/c', 'b/d/e'])
+        self.assertCmpPathByDirblock(['kl', 'ab/cd', 'ab/ef', 'gh/ij'])
+
+    def test_tricky_paths(self):
+        self.assertCmpPathByDirblock([
+            # Contents of ''
+            '', 'a', 'a-a', 'a=a', 'b',
+            # Contents of 'a'
+            'a/a', 'a/a-a', 'a/a=a', 'a/b',
+            # Contents of 'a/a'
+            'a/a/a', 'a/a/a-a', 'a/a/a=a',
+            # Contents of 'a/a/a'
+            'a/a/a/a', 'a/a/a/b',
+            # Contents of 'a/a/a-a',
+            'a/a/a-a/a', 'a/a/a-a/b',
+            # Contents of 'a/a/a=a',
+            'a/a/a=a/a', 'a/a/a=a/b',
+            # Contents of 'a/a-a'
+            'a/a-a/a',
+            # Contents of 'a/a-a/a'
+            'a/a-a/a/a', 'a/a-a/a/b',
+            # Contents of 'a/a=a'
+            'a/a=a/a',
+            # Contents of 'a/b'
+            'a/b/a', 'a/b/b',
+            # Contents of 'a-a',
+            'a-a/a', 'a-a/b',
+            # Contents of 'a=a',
+            'a=a/a', 'a=a/b',
+            # Contents of 'b',
+            'b/a', 'b/b',
+            ])
+        self.assertCmpPathByDirblock([
+                 # content of '/'
+                 '', 'a', 'a-a', 'a-z', 'a=a', 'a=z',
+                 # content of 'a/'
+                 'a/a', 'a/a-a', 'a/a-z',
+                 'a/a=a', 'a/a=z',
+                 'a/z', 'a/z-a', 'a/z-z',
+                 'a/z=a', 'a/z=z',
+                 # content of 'a/a/'
+                 'a/a/a', 'a/a/z',
+                 # content of 'a/a-a'
+                 'a/a-a/a',
+                 # content of 'a/a-z'
+                 'a/a-z/z',
+                 # content of 'a/a=a'
+                 'a/a=a/a',
+                 # content of 'a/a=z'
+                 'a/a=z/z',
+                 # content of 'a/z/'
+                 'a/z/a', 'a/z/z',
+                 # content of 'a-a'
+                 'a-a/a',
+                 # content of 'a-z'
+                 'a-z/z',
+                 # content of 'a=a'
+                 'a=a/a',
+                 # content of 'a=z'
+                 'a=z/z',
+                ])
+
+
+class TestCompiledCmpPathByDirblock(TestCmpPathByDirblock):
+    """Test the pyrex implementation of cmp_path_by_dirblock"""
+
+    _test_needs_features = [CompiledDirstateHelpersFeature]
+
+    def get_cmp_by_dirs(self):
+        from bzrlib._dirstate_helpers_c import cmp_path_by_dirblock_c
+        return cmp_path_by_dirblock_c
+
+
+class TestMemRChr(tests.TestCase):
+    """Test the pyrex implementation of cmp_path_by_dirblock"""
+
+    _test_needs_features = [CompiledDirstateHelpersFeature]
+
+    def assertMemRChr(self, expected, s, c):
+        from bzrlib._dirstate_helpers_c import _py_memrchr
+        self.assertEqual(expected, _py_memrchr(s, c))
+
+    def test_missing(self):
+        self.assertMemRChr(None, '', 'a')
+        self.assertMemRChr(None, '', 'c')
+        self.assertMemRChr(None, 'abcdefghijklm', 'q')
+        self.assertMemRChr(None, 'aaaaaaaaaaaaaaaaaaaaaaa', 'b')
+
+    def test_single_entry(self):
+        self.assertMemRChr(0, 'abcdefghijklm', 'a')
+        self.assertMemRChr(1, 'abcdefghijklm', 'b')
+        self.assertMemRChr(2, 'abcdefghijklm', 'c')
+        self.assertMemRChr(10, 'abcdefghijklm', 'k')
+        self.assertMemRChr(11, 'abcdefghijklm', 'l')
+        self.assertMemRChr(12, 'abcdefghijklm', 'm')
+
+    def test_multiple(self):
+        self.assertMemRChr(10, 'abcdefjklmabcdefghijklm', 'a')
+        self.assertMemRChr(11, 'abcdefjklmabcdefghijklm', 'b')
+        self.assertMemRChr(12, 'abcdefjklmabcdefghijklm', 'c')
+        self.assertMemRChr(20, 'abcdefjklmabcdefghijklm', 'k')
+        self.assertMemRChr(21, 'abcdefjklmabcdefghijklm', 'l')
+        self.assertMemRChr(22, 'abcdefjklmabcdefghijklm', 'm')
+        self.assertMemRChr(22, 'aaaaaaaaaaaaaaaaaaaaaaa', 'a')
+
+    def test_with_nulls(self):
+        self.assertMemRChr(10, 'abc\0\0\0jklmabc\0\0\0ghijklm', 'a')
+        self.assertMemRChr(11, 'abc\0\0\0jklmabc\0\0\0ghijklm', 'b')
+        self.assertMemRChr(12, 'abc\0\0\0jklmabc\0\0\0ghijklm', 'c')
+        self.assertMemRChr(20, 'abc\0\0\0jklmabc\0\0\0ghijklm', 'k')
+        self.assertMemRChr(21, 'abc\0\0\0jklmabc\0\0\0ghijklm', 'l')
+        self.assertMemRChr(22, 'abc\0\0\0jklmabc\0\0\0ghijklm', 'm')
+        self.assertMemRChr(22, 'aaa\0\0\0aaaaaaa\0\0\0aaaaaaa', 'a')
+        self.assertMemRChr(9, '\0\0\0\0\0\0\0\0\0\0', '\0')

=== modified file 'bzrlib/tests/test_dirstate.py'
--- bzrlib/tests/test_dirstate.py	2007-07-10 14:51:23 +0000
+++ bzrlib/tests/test_dirstate.py	2007-07-11 00:01:54 +0000
@@ -189,11 +189,12 @@
               c
               d/
                 e
+            b-c
             f
         """
         tree = self.make_branch_and_tree('tree')
-        paths = ['a', 'b/', 'b/c', 'b/d/', 'b/d/e', 'f']
-        file_ids = ['a-id', 'b-id', 'c-id', 'd-id', 'e-id', 'f-id']
+        paths = ['a', 'b/', 'b/c', 'b/d/', 'b/d/e', 'b-c', 'f']
+        file_ids = ['a-id', 'b-id', 'c-id', 'd-id', 'e-id', 'b-c-id', 'f-id']
         self.build_tree(['tree/' + p for p in paths])
         tree.set_root_id('TREE_ROOT')
         tree.add([p.rstrip('/') for p in paths], file_ids)
@@ -214,6 +215,9 @@
         e_text = t.get_bytes('b/d/e')
         e_sha = osutils.sha_string(e_text)
         e_len = len(e_text)
+        b_c_text = t.get_bytes('b-c')
+        b_c_sha = osutils.sha_string(b_c_text)
+        b_c_len = len(b_c_text)
         # f_packed_stat = dirstate.pack_stat(os.stat('tree/f'))
         f_text = t.get_bytes('f')
         f_sha = osutils.sha_string(f_text)
@@ -244,6 +248,10 @@
                       ('f', '', 0, False, null_stat),
                       ('f', e_sha, e_len, False, revision_id),
                      ]),
+            'b-c':(('', 'b-c', 'b-c-id'), [
+                      ('f', '', 0, False, null_stat),
+                      ('f', b_c_sha, b_c_len, False, revision_id),
+                     ]),
             'f':(('', 'f', 'f-id'), [
                   ('f', '', 0, False, null_stat),
                   ('f', f_sha, f_len, False, revision_id),
@@ -276,12 +284,12 @@
         tree, state, expected = self.create_basic_dirstate()
         # Now we will just remove and add every file so we get an extra entry
         # per entry. Unversion in reverse order so we handle subdirs
-        tree.unversion(['f-id', 'e-id', 'd-id', 'c-id', 'b-id', 'a-id'])
-        tree.add(['a', 'b', 'b/c', 'b/d', 'b/d/e', 'f'],
-                 ['a-id2', 'b-id2', 'c-id2', 'd-id2', 'e-id2', 'f-id2'])
+        tree.unversion(['f-id', 'b-c-id', 'e-id', 'd-id', 'c-id', 'b-id', 'a-id'])
+        tree.add(['a', 'b', 'b/c', 'b/d', 'b/d/e', 'b-c', 'f'],
+                 ['a-id2', 'b-id2', 'c-id2', 'd-id2', 'e-id2', 'b-c-id2', 'f-id2'])
 
         # Update the expected dictionary.
-        for path in ['a', 'b', 'b/c', 'b/d', 'b/d/e', 'f']:
+        for path in ['a', 'b', 'b/c', 'b/d', 'b/d/e', 'b-c', 'f']:
             orig = expected[path]
             path2 = path + '2'
             # This record was deleted in the current tree
@@ -344,6 +352,7 @@
             state.lock_read()
         return tree, state, expected
 
+
 class TestTreeToDirState(TestCaseWithDirState):
 
     def test_empty_to_dirstate(self):
@@ -1750,23 +1759,24 @@
                       (dir, name) tuples, and sorted according to how _bisect
                       requires.
         """
-        dir_names = sorted(osutils.split(p) for p in paths)
-        result = state._bisect(dir_names)
+        result = state._bisect(paths)
         # For now, results are just returned in whatever order we read them.
         # We could sort by (dir, name, file_id) or something like that, but in
         # the end it would still be fairly arbitrary, and we don't want the
         # extra overhead if we can avoid it. So sort everything to make sure
         # equality is true
-        assert len(map_keys) == len(dir_names)
+        assert len(map_keys) == len(paths)
         expected = {}
-        for dir_name, keys in zip(dir_names, map_keys):
+        for path, keys in zip(paths, map_keys):
             if keys is None:
                 # This should not be present in the output
                 continue
-            expected[dir_name] = sorted(expected_map[k] for k in keys)
+            expected[path] = sorted(expected_map[k] for k in keys)
 
-        for dir_name in result:
-            result[dir_name].sort()
+        # The returned values are just arranged randomly based on when they
+        # were read, for testing, make sure it is properly sorted.
+        for path in result:
+            result[path].sort()
 
         self.assertEqual(expected, result)
 
@@ -1809,8 +1819,7 @@
             dir_name_id, trees_info = entry
             expected[dir_name_id] = trees_info
 
-        dir_names = sorted(osutils.split(p) for p in paths)
-        result = state._bisect_recursive(dir_names)
+        result = state._bisect_recursive(paths)
 
         self.assertEqual(expected, result)
 
@@ -1825,6 +1834,7 @@
         self.assertBisect(expected, [['b/c']], state, ['b/c'])
         self.assertBisect(expected, [['b/d']], state, ['b/d'])
         self.assertBisect(expected, [['b/d/e']], state, ['b/d/e'])
+        self.assertBisect(expected, [['b-c']], state, ['b-c'])
         self.assertBisect(expected, [['f']], state, ['f'])
 
     def test_bisect_multi(self):
@@ -1836,6 +1846,8 @@
         # ('', 'f') sorts before the others
         self.assertBisect(expected, [['f'], ['b/d'], ['b/d/e']],
                           state, ['b/d', 'b/d/e', 'f'])
+        self.assertBisect(expected, [['b'], ['b/c'], ['b-c']],
+                          state, ['b', 'b/c', 'b-c'])
 
     def test_bisect_one_page(self):
         """Test bisect when there is only 1 page to read"""
@@ -1847,12 +1859,15 @@
         self.assertBisect(expected,[['b/c']], state, ['b/c'])
         self.assertBisect(expected,[['b/d']], state, ['b/d'])
         self.assertBisect(expected,[['b/d/e']], state, ['b/d/e'])
+        self.assertBisect(expected,[['b-c']], state, ['b-c'])
         self.assertBisect(expected,[['f']], state, ['f'])
         self.assertBisect(expected,[['a'], ['b'], ['f']],
                           state, ['a', 'b', 'f'])
         # ('', 'f') sorts before the others
         self.assertBisect(expected, [['f'], ['b/d'], ['b/d/e']],
                           state, ['b/d', 'b/d/e', 'f'])
+        self.assertBisect(expected, [['b'], ['b/c'], ['b-c']],
+                          state, ['b', 'b/c', 'b-c'])
 
     def test_bisect_duplicate_paths(self):
         """When bisecting for a path, handle multiple entries."""
@@ -1866,6 +1881,7 @@
         self.assertBisect(expected, [['b/d', 'b/d2']], state, ['b/d'])
         self.assertBisect(expected, [['b/d/e', 'b/d/e2']],
                           state, ['b/d/e'])
+        self.assertBisect(expected, [['b-c', 'b-c2']], state, ['b-c'])
         self.assertBisect(expected, [['f', 'f2']], state, ['f'])
 
     def test_bisect_page_size_too_small(self):
@@ -1878,6 +1894,7 @@
         self.assertBisect(expected, [['b/c']], state, ['b/c'])
         self.assertBisect(expected, [['b/d']], state, ['b/d'])
         self.assertBisect(expected, [['b/d/e']], state, ['b/d/e'])
+        self.assertBisect(expected, [['b-c']], state, ['b-c'])
         self.assertBisect(expected, [['f']], state, ['f'])
 
     def test_bisect_missing(self):
@@ -1886,6 +1903,7 @@
         self.assertBisect(expected, [None], state, ['foo'])
         self.assertBisect(expected, [None], state, ['b/foo'])
         self.assertBisect(expected, [None], state, ['bar/foo'])
+        self.assertBisect(expected, [None], state, ['b-c/foo'])
 
         self.assertBisect(expected, [['a'], None, ['b/d']],
                           state, ['a', 'foo', 'b/d'])
@@ -1907,13 +1925,14 @@
     def test_bisect_dirblocks(self):
         tree, state, expected = self.create_duplicated_dirstate()
         self.assertBisectDirBlocks(expected,
-            [['', 'a', 'a2', 'b', 'b2', 'f', 'f2']], state, [''])
+            [['', 'a', 'a2', 'b', 'b2', 'b-c', 'b-c2', 'f', 'f2']],
+            state, [''])
         self.assertBisectDirBlocks(expected,
             [['b/c', 'b/c2', 'b/d', 'b/d2']], state, ['b'])
         self.assertBisectDirBlocks(expected,
             [['b/d/e', 'b/d/e2']], state, ['b/d'])
         self.assertBisectDirBlocks(expected,
-            [['', 'a', 'a2', 'b', 'b2', 'f', 'f2'],
+            [['', 'a', 'a2', 'b', 'b2', 'b-c', 'b-c2', 'f', 'f2'],
              ['b/c', 'b/c2', 'b/d', 'b/d2'],
              ['b/d/e', 'b/d/e2'],
             ], state, ['', 'b', 'b/d'])
@@ -1934,6 +1953,7 @@
         self.assertBisectRecursive(expected, ['a'], state, ['a'])
         self.assertBisectRecursive(expected, ['b/c'], state, ['b/c'])
         self.assertBisectRecursive(expected, ['b/d/e'], state, ['b/d/e'])
+        self.assertBisectRecursive(expected, ['b-c'], state, ['b-c'])
         self.assertBisectRecursive(expected, ['b/d', 'b/d/e'],
                                    state, ['b/d'])
         self.assertBisectRecursive(expected, ['b', 'b/c', 'b/d', 'b/d/e'],
@@ -2024,4 +2044,3 @@
         self.assertContainsRe(str(e),
             'file a-id is absent in row')
 
-

